
#
# This file is part of the ATC-pie project,
# an air traffic control simulation program.
# 
# Copyright (C) 2015  Michael Filhol <mickybadia@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
#

from datetime import timedelta

from PyQt5.QtWidgets import QWidget, QMenu, QAction, QActionGroup, QMessageBox
from PyQt5.QtGui import QIcon
from PyQt5.QtCore import QSortFilterProxyModel
from ui.fplPanel import Ui_FplPanel

from base.util import some
from base.fpl import FPL
from base.nav import world_navpoint_db, NavpointError
from base.utc import now, datestr, timestr

from session.config import settings
from session.env import env
from session.manager import OnlineFplActionBlocked

from gui.misc import signals, selection, IconFile
from gui.dialog.detailSheets import FPLdetailSheetDialog
from gui.dialog.miscDialogs import yesNo_question


# ---------- Constants ----------

filter_includes_missing_dates = True

# -------------------------------




def AD_on_map(icao):
	try:
		return env.pointOnMap(world_navpoint_db.findAirfield(icao).coordinates)
	except NavpointError:
		return False



acceptAll = lambda x: True


def ckArrDepAltAD(fpl, f):
	return fpl[FPL.ICAO_DEP] is not None and f(fpl[FPL.ICAO_DEP]) \
			or fpl[FPL.ICAO_ARR] is not None and f(fpl[FPL.ICAO_ARR]) \
			or fpl[FPL.ICAO_ALT] is not None and f(fpl[FPL.ICAO_ALT])




class FplFilterModel(QSortFilterProxyModel):
	def __init__(self, base_model, parent=None):
		QSortFilterProxyModel.__init__(self, parent)
		self.setSourceModel(base_model)
		self.accept_closed = True
		self.accept_outdated = True
		self.callsign_filter = acceptAll
		self.arrDep_filter = acceptAll
		self.date_filter = acceptAll

	def filterAcceptsRow(self, sourceRow, sourceParent):
		fpl = self.sourceModel().FPL_list[sourceRow]
		return (self.accept_closed or fpl.onlineStatus() != FPL.CLOSED) \
				and (self.accept_outdated or not fpl.isOutdated()) \
				and self.callsign_filter(fpl) and self.arrDep_filter(fpl) \
				and (filter_includes_missing_dates if fpl[FPL.TIME_OF_DEP] is None else self.date_filter(fpl))
	
	def filter_statusToggle_closed(self, toggle):
		self.accept_closed = toggle
		self.invalidateFilter()
	
	def filter_statusToggle_outdated(self, toggle):
		self.accept_outdated = toggle
		self.invalidateFilter()
	
	def filter_callsign(self, fstring):
		lower = fstring.lower()
		self.callsign_filter = lambda fpl: lower in some(fpl[FPL.CALLSIGN], '').lower()
		self.invalidateFilter()
	
	def filter_date_today(self):
		self.date_filter = lambda fpl: fpl.flightIsInTimeWindow(timedelta(hours=24))
		self.invalidateFilter()
	
	def filter_date_week(self):
		self.date_filter = lambda fpl: fpl.flightIsInTimeWindow(timedelta(days=7))
		self.invalidateFilter()
	
	def filter_arrDep_all(self):
		self.arrDep_filter = acceptAll
		self.invalidateFilter()
	
	def filter_arrDep_inRange(self):
		self.arrDep_filter = lambda fpl: ckArrDepAltAD(fpl, AD_on_map)
		self.invalidateFilter()
	
	def filter_arrDep_here(self):
		here = settings.location_code
		self.arrDep_filter = lambda fpl: ckArrDepAltAD(fpl, (lambda ad: ad is not None and ad.upper() == here))
		self.invalidateFilter()




# ================================================ #

#                     WIDGETS                      #

# ================================================ #

class FlightPlansPanel(QWidget, Ui_FplPanel):
	def __init__(self, parent=None):
		QWidget.__init__(self, parent)
		self.setupUi(self)
		self.newFPL_button.setIcon(QIcon(IconFile.action_newFPL))
		self.remove_button.setIcon(QIcon(IconFile.button_bin))
		self.list_model = FplFilterModel(env.FPLs, self)
		self.list_view.setModel(self.list_model)
		self.list_view.horizontalHeader().resizeSection(0, 40)
		self.filterCallsign_edit.setClearButtonEnabled(True)
		# Status filters
		status_filter_menu = QMenu()
		status_filter_menu.addAction(self._filter_toggle('Show closed', self.list_model.filter_statusToggle_closed))
		status_filter_menu.addAction(self._filter_toggle('Show outdated', self.list_model.filter_statusToggle_outdated))
		self.filterStatus_button.setMenu(status_filter_menu)
		# Airport filters
		arrDep_action_group = QActionGroup(self)
		arrDep_action_group.addAction(self._filter_action('All', self.list_model.filter_arrDep_all, ticked=True))
		arrDep_action_group.addAction(self._filter_action('On map', self.list_model.filter_arrDep_inRange))
		if env.airport_data is not None:
			arrDep_action_group.addAction(self._filter_action('Here only', self.list_model.filter_arrDep_here))
		arrDep_filter_menu = QMenu()
		arrDep_filter_menu.addActions(arrDep_action_group.actions())
		self.filterArrDep_button.setMenu(arrDep_filter_menu)
		# Date filters
		date_action_group = QActionGroup(self)
		date_action_group.addAction(self._filter_action('+/- 3 days', self.list_model.filter_date_week))
		date_action_group.addAction(self._filter_action('Today (~24 h)', self.list_model.filter_date_today, ticked=True))
		date_filter_menu = QMenu()
		date_filter_menu.addActions(date_action_group.actions())
		self.filterDate_button.setMenu(date_filter_menu)
		# Finish up
		self.updateButtons()
		self.filterCallsign_edit.textChanged.connect(self.list_model.filter_callsign)
		self.checkNow_button.clicked.connect(self.syncOnlineNow)
		self.publish_button.clicked.connect(self.publishFplOnline)
		self.open_button.clicked.connect(self.openFplOnline)
		self.close_button.clicked.connect(self.closeFplOnline)
		self.newFPL_button.clicked.connect(lambda: self.createLocalFPL(link=None))
		self.revert_button.clicked.connect(self.revertLocalChanges)
		self.remove_button.clicked.connect(self.removeFPL)
		signals.selectionChanged.connect(self.updateButtons)
		signals.fplEditRequest.connect(self.editFPL)
		signals.newLinkedFPLrequest.connect(lambda s: self.createLocalFPL(link=s))
		signals.sessionStarted.connect(self.updateButtons)
		signals.sessionEnded.connect(self.updateButtons)
		
	def _filter_toggle(self, text, toggle_function, ticked=False):
		action = QAction(text, self)
		action.setCheckable(True)
		action.setChecked(ticked)
		toggle_function(ticked)
		action.toggled.connect(toggle_function)
		return action
	
	def _filter_action(self, text, f, ticked=False):
		action = QAction(text, self)
		action.setCheckable(True)
		action.setChecked(ticked)
		action.triggered.connect(f)
		if ticked:
			f()
		return action
	
	def updateButtons(self):
		fpl = selection.fpl
		self.publish_button.setEnabled(fpl is not None and (not fpl.isOnline() or fpl.hasLocalChanges()))
		self.open_button.setEnabled(fpl is not None and fpl.onlineStatus() == FPL.FILED)
		self.close_button.setEnabled(fpl is not None and fpl.onlineStatus() == FPL.OPEN)
		self.revert_button.setEnabled(fpl is not None and fpl.hasLocalChanges())
		self.remove_button.setEnabled(fpl is not None and not fpl.isOnline() and env.linkedStrip(fpl) is None)
	
	def createLocalFPL(self, link=None):
		"""
		Optional strip to link to after FPL is created. If given, it must already be in the live strips model.
		"""
		new_fpl = FPL()
		if link is not None:
			for d in FPL.details:
				new_fpl[d] = link.lookup(d, fpl=False)
		dialog = FPLdetailSheetDialog(self, new_fpl)
		dialog.exec()
		if dialog.result() > 0: # not rejected
			env.FPLs.addFPL(new_fpl)
			if link is not None:
				link.linkFPL(new_fpl, autoFillOK=False)
			selection.selectFPL(new_fpl)
	
	def editFPL(self, fpl):
		FPLdetailSheetDialog(self, fpl).exec()
		self.updateButtons()
	
	def publishFplOnline(self):
		if settings.session_manager.isRunning():
			fpl = selection.fpl
			if fpl is not None:
				try:
					settings.session_manager.pushFplOnline(fpl)
				except OnlineFplActionBlocked as err:
					QMessageBox.critical(self, 'FPL upload error', str(err))
				self.updateButtons()
	
	def syncOnlineNow(self):
		if settings.session_manager.isRunning():
			try:
				settings.session_manager.syncOnlineFPLs()
			except OnlineFplActionBlocked as err:
				QMessageBox.critical(self, 'FPL sync error', str(err))
	
	def openFplOnline(self):
		if settings.session_manager.isRunning():
			fpl = selection.fpl
			if fpl is not None:
				tnow = now()
				text = 'Update DEP time with current UTC date & time below before opening FPL?\n' \
						'%s, %s' % (datestr(tnow), timestr(tnow))
				button = QMessageBox.question(self, 'Open FPL', text, buttons=(QMessageBox.Cancel | QMessageBox.No | QMessageBox.Yes))
				if button == QMessageBox.Yes:
					fpl[FPL.TIME_OF_DEP] = tnow
				if button != QMessageBox.Cancel:
					try:
						settings.session_manager.changeFplStatus(fpl, FPL.OPEN)
					except OnlineFplActionBlocked as err:
						QMessageBox.critical(self, 'FPL open/close error', str(err))
				self.updateButtons()
	
	def closeFplOnline(self):
		if settings.session_manager.isRunning():
			fpl = selection.fpl
			if fpl is not None and yesNo_question(self, 'Close FPL',
					'Time is %s (UTC).' % timestr(now()), 'Close the flight plan online?'):
				try:
					settings.session_manager.changeFplStatus(fpl, FPL.CLOSED)
				except OnlineFplActionBlocked as err:
					QMessageBox.critical(self, 'FPL open/close error', str(err))
				self.updateButtons()
	
	def revertLocalChanges(self):
		if selection.fpl is not None:
			selection.fpl.revertToOnlineValues()
			env.FPLs.refreshViews()
			self.updateButtons()
	
	def removeFPL(self):
		if selection.fpl is not None:
			env.FPLs.removeFPL(selection.fpl)
			selection.deselect()
			self.updateButtons()
